#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ====================================================
# Copyright 2018 Christopher Simpkins
# MIT License
# ====================================================

"""
The app.py module defines a main() function that includes the logic for the `ufodiff` command line executable.

This command line executable provides file modification (add, delete, modified) and text diff reports
for files and directory structure that are part of the Unified Font Object (UFO) specification.  These source
files are used in typeface development.
"""

import os
import sys
from commandlines import Command
from standardstreams import stdout, stderr

from ufodiff import (
    settings,
)  # defines application version, help string, version string, usage string
from ufodiff.subcommands.delta import Delta
from ufodiff.subcommands.diff import Diff

from ufodiff.utilities import dir_exists


def main():
    """Defines the logic for the `ufodiff` command line executable"""
    c = Command()

    if c.does_not_validate_missing_args():
        stderr(
            "[ufodiff] ERROR: Please include the appropriate arguments with your command."
        )
        sys.exit(1)

    if c.is_help_request():
        stdout(settings.HELP)
        sys.exit(0)
    elif c.is_version_request():
        stdout(settings.VERSION)
        sys.exit(0)
    elif c.is_usage_request():
        stdout(settings.USAGE)
        sys.exit(0)

    # DELTA + DELTAJSON + DELTAMD sub-commands
    if c.subcmd in {"delta", "deltajson", "deltamd"}:
        # argument validation
        validate_delta_commands_args(c)
        # create list for UFO filtered analyses as requested by user
        ufo_directory_list = []
        for arg in c.argv:
            if arg.endswith(".ufo"):
                ufo_directory_list.append(arg)

        # flags for type of test
        is_branch_test = False
        is_commits_test = False

        if c.arg2.startswith("commits:"):
            is_commits_test = True
            commits_list = c.arg2.split(":")
            commit_number = commits_list[1]
        elif c.arg2.startswith("branch:"):
            is_branch_test = True
            branch_list = c.arg2.split(":")
            branch_name = branch_list[1]
        # else:
        #     pass  # TODO: add direct git idiom call support

        # recursive search for the root of the git repository x 3 levels if not found in working directory
        verified_gitroot_path = get_git_root_path()

        # perform the delta analysis on the repository, different object for commits vs branch tests
        if is_commits_test is True:
            delta = Delta(
                verified_gitroot_path,
                ufo_directory_list,
                is_commit_test=True,
                commit_number=commit_number,
            )
        elif is_branch_test is True:
            delta = Delta(
                verified_gitroot_path,
                ufo_directory_list,
                is_branch_test=True,
                compare_branch_name=branch_name,
            )

        if c.arg1 == "all":
            if c.subcmd == "delta":
                sys.stdout.write(delta.get_stdout_string(write_format="text"))
            elif c.subcmd == "deltajson":
                sys.stdout.write(delta.get_stdout_string(write_format="json"))
            elif c.subcmd == "deltamd":
                sys.stdout.write(delta.get_stdout_string(write_format="markdown"))
        # elif c.arg1 == "glyph":
        #     pass  # TODO: implement glyph only command handling with 'ufo delta glyph'
        # elif c.arg1 == "nonglyph":
        #     pass  # TODO: implement nonglyph only command handling with 'ufo delta nonglyph'
    # DIFF SUBCOMMAND
    elif c.subcmd == "diff":
        # argument validation
        validate_diff_commands_args(c)
        # execute the command
        try:
            verified_gitroot_path = get_git_root_path()
            diff = Diff(verified_gitroot_path, color_diff=True)
            for diff_string in diff.get_diff_string_generator(c.arg1):
                stdout(diff_string)
        except Exception as e:
            stderr(
                "[ufodiff] ERROR: Unable to excecute your request. Error returned as: "
                + os.linesep
                + str(e)
            )
            sys.exit(1)
    # DIFFNC SUBCOMMAND
    elif c.subcmd == "diffnc":
        # argument validations
        validate_diff_commands_args(c)
        # execute the command
        try:
            verified_gitroot_path = get_git_root_path()
            diff = Diff(verified_gitroot_path, color_diff=False)
            for diff_string in diff.get_diff_string_generator(c.arg1):
                stdout(diff_string)
        except Exception as e:
            stderr(
                "[ufodiff] ERROR: Unable to excecute your request. Error returned as: "
                + os.linesep
                + str(e)
            )
            sys.exit(1)
    # # DIFF-FILE SUBCOMMAND
    # elif c.subcmd == "diff-filter":  # user specified file/directory filters on the diff performed
    #     pass
    # # DIFF-FILENC SUBCOMMAND
    # elif c.subcmd == "diff-filternc":
    #     pass
    sys.exit(0)


# Command Line Utility Functions


def validate_delta_commands_args(command_obj):
    """
    Validates arguments for any delta/deltajson/deltamd command requested by user.  It provides
    user error messages and raises SystemExit for erroneous command entry at the command line.

    :param command_obj: a commandlines library Command object
    :return: no return object, SystemExit raised for all errors detected
    """
    acceptable_deltacommands = ["all"]  # used in command line argument validations
    # Command line argument validations
    if command_obj.argc < 3:  # expected argument number
        stderr("[ufodiff] ERROR: Missing arguments.")
        sys.exit(1)
    if (
        command_obj.arg1 not in acceptable_deltacommands
    ):  # acceptable sub command to delta
        stderr(
            "[ufodiff] ERROR: 'ufodiff "
            + command_obj.arg0
            + " "
            + command_obj.arg1
            + "' is not a valid request"
        )
        stderr("Acceptable arguments to " + command_obj.arg0 + " include:")
        for acceptable_deltacommand in acceptable_deltacommands:
            stderr(" " + acceptable_deltacommand)
        sys.exit(1)
    if command_obj.arg2.startswith("commits:"):
        commits_list = command_obj.arg2.split(":")
        if (
            len(commits_list) == 2 and commits_list[1] == ""
        ):  # does not include a value following the colon
            stderr(
                "[ufodiff] ERROR: Please include an integer value following the 'commits:' argument"
            )
            sys.exit(1)
        else:
            commits_number = commits_list[1]
            validate_commit_number(commits_number)
    elif command_obj.arg2.startswith("branch:"):
        branch_list = command_obj.arg2.split(":")
        if len(branch_list) == 2 and branch_list[1] == "":
            stderr(
                "[ufodiff] ERROR: Please include the name of an existing git branch following the 'branch:' argument"
            )
            sys.exit(1)
    else:
        stderr(
            "[ufodiff] ERROR: Please include either the 'commits:' or 'branch:' argument in the command"
        )
        sys.exit(1)


def validate_diff_commands_args(command_obj):
    """
    Validates arguments for any diff/diffnc command requested by user.  It provides user error
    messages and raises SystemExit for erroneous command entry at the command line.

    :param command_obj: a commandlines library Command object
    :return: no return object, SystemExit raised for all errors detected
    """
    # argument validations
    if command_obj.argc < 2:
        stderr("[ufodiff] ERROR: Missing arguments.")
        sys.exit(1)
    if command_obj.arg1.startswith("commits:"):
        if len(command_obj.arg1) < 9:
            stderr(
                "[ufodiff] ERROR: Please include an integer after the colon in the 'commits:[number]' argument"
            )
            sys.exit(1)
        else:
            commits_list = command_obj.arg1.split(":")
            commits_number = commits_list[1]
            # validate the number of commits as an integer value, exit with error message if not an integer
            validate_commit_number(commits_number)
    if command_obj.arg1.startswith("branch:"):
        if len(command_obj.arg1) < 8:
            stderr(
                "[ufodiff] ERROR: Please include the name of a git branch following the colon in the "
                "'branch:[name]' argument"
            )
            sys.exit(1)


def validate_commit_number(commits_number):
    """
    Validates commit number entered by user following `commits:` argument as valid integer and within appropriate range

    :param commits_number: string that was entered by user following `commits:` argument
    :return: no return object, raises SystemExit for all errors detected
    """
    if (
        not commits_number.isdigit()
    ):  # validate that user entered number of commits for diff is an integer
        stderr(
            "[ufodiff] ERROR: The value following the colon in the 'commits:[number]' argument is not a valid "
            "integer value"
        )
        sys.exit(1)
    elif (
        int(commits_number) == 0 or int(commits_number) < 0
    ):  # validate that the number of commits is > 0
        stderr(
            "[ufodiff] ERROR: Please define a value over zero for the commit history"
        )
        sys.exit(1)


def get_git_root_path():
    """
    Recursively searches for git root path over 4 directory levels above working directory
    :return: validated git root path as string OR raises SystemExit if not found
    """
    try:
        # begin by defining current working directory as root of git repository
        unverified_gitroot_path = os.path.abspath(".")

        # check to see if this assumption is correct
        if dir_exists(os.path.join(unverified_gitroot_path, ".git")):
            verified_gitroot_path = os.path.join(unverified_gitroot_path, ".git")
        else:  # if not, recursive search up to three directories above for the git repo root
            one_level_up = os.path.abspath(
                os.path.join(unverified_gitroot_path, os.pardir)
            )
            two_levels_up = os.path.dirname(one_level_up)
            three_levels_up = os.path.dirname(two_levels_up)

            one_level_up_path = os.path.join(one_level_up, ".git")
            two_levels_up_path = os.path.join(two_levels_up, ".git")
            three_levels_up_path = os.path.join(three_levels_up, ".git")

            if dir_exists(one_level_up_path):  # check one directory level up
                verified_gitroot_path = os.path.dirname(one_level_up_path)
            elif dir_exists(two_levels_up_path):  # check two directory levels up
                verified_gitroot_path = os.path.dirname(two_levels_up_path)
            elif dir_exists(three_levels_up_path):  # check three directory levels up
                verified_gitroot_path = os.path.dirname(three_levels_up_path)
            else:
                stderr(
                    "[ufodiff] ERROR: Unable to identify the root of your git repository. Please try again from "
                    "the root of your repository"
                )
                sys.exit(1)
    except Exception as e:
        stderr(
            "[ufodiff] ERROR: Unable to identify the root of your git repository. Please try again from "
            "the root of your repository. " + str(e)
        )
        sys.exit(1)

    return verified_gitroot_path



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ====================================================
# Copyright 2017 Christopher Simpkins
# MIT License
# ====================================================

import os


class Ufo(object):
    """
    Validates filepaths based upon the Unified Font Object source specification
    """

    def __init__(self):
        self.acceptable_files = {
            "metainfo.plist",
            "fontinfo.plist",
            "groups.plist",
            "kerning.plist",
            "features.fea",
            "lib.plist",
            "contents.plist",
            "layercontents.plist",
            "layerinfo.plist",
        }

    def _is_images_directory_file(self, filepath):
        path_list = filepath.split(os.path.sep)
        index = 0
        for path_part in path_list:
            if path_part == "images":
                ufo_test_directory = path_list[
                    index - 1
                ]  # .ufo directory should be one directory level above images
                base_filename = path_list[
                    -1
                ]  # base filename should have .png extension
                if (
                    len(base_filename) > 4
                    and base_filename[-4:] == ".png"
                    and ufo_test_directory[-4:] == ".ufo"
                ):
                    return True  # .png file in *.ufo/images directory = pass
            else:
                index += 1
        return False  # if iterate through entire path and never satisfy above test conditions, test fails

    def _is_data_directory_file(self, filepath):
        path_list = filepath.split(os.path.sep)
        index = 0
        for path_part in path_list:
            if path_part == "data":
                ufo_test_directory = path_list[
                    index - 1
                ]  # .ufo directory should be one level above data
                if (
                    ufo_test_directory[-4:] == ".ufo"
                ):  # test for presence of *.ufo directory one level up
                    return True  # any file in *.ufo/data directory = pass
            else:
                index += 1
        return False  # if iterate through entire path and never satisfy above test conditions, test fails

    def validate_file(self, filepath):
        if (
            os.path.basename(filepath) in self.acceptable_files
            or filepath[-5:] == ".glif"
        ):
            return True
        elif (
            self._is_data_directory_file(filepath) is True
        ):  # UFO v3 data directory file test
            return True
        elif (
            self._is_images_directory_file(filepath) is True
        ):  # UFO v3 images directory file test
            return True
        else:
            return False

    def get_valid_file_filterlist_for_diff(self):
        filter_list = []
        # add valid non-glyph file wildcards
        for a_file in self.acceptable_files:
            filter_string = "*" + a_file
            filter_list.append(filter_string)
        # add images directory wildcard (added in UFO v3)
        filter_list.append("*\.ufo/images/*")
        # add data directory wildcard (added in UFO v3)
        filter_list.append("*\.ufo/data/*")
        # add glyph file filters
        filter_list.append("*.glif")
        return filter_list

    def is_nonglyph_file(self, filepath):
        if filepath[-6:] == ".plist" or filepath[-4:] == ".fea":
            return True
        else:
            return False

    def is_glyph_file(self, filepath):
        if filepath[-5:] == ".glif":
            return True
        else:
            return False

    def is_ufo_version_file(self, filepath):
        if os.path.basename(filepath) == "metainfo.plist":
            return True
        else:
            return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ====================================================
# Copyright 2018 Christopher Simpkins
# MIT License
# ====================================================

import os
import json

from git import Repo

from ufodiff.utilities.ufo import Ufo
from ufodiff.settings import major_version, minor_version, patch_version


class Delta(object):
    """
    Delta class stores delta subcommand data and provides methods to support creation of
    delta / deltajson / deltamd application subcommand reports through UFO file spec validation and
    filtering of user requested *.ufo source directory filters

    Uses DeltaFilepathStringDict object (this module) to (1) filter data by optional user specified *.ufo source filter;
    (2) maintain final data for reports in the DeltaFilepathStringDict.delta_dict Python dictionary property

    :param gitrepo_path: (string) absolute file path to the root level of the git repository for analysis (automatically detected in ufodiff.app.py)
    :param ufo_directory_list: (list) list of one or more UFO directories for filter of results (user specified on CL)
    :param is_commit_test: (boolean) flag for request as test of commit history in the git repository
    :param commit_number: (string) the number of commits in git commit history for analysis as string (user specified)
    :param is_branch_test: (boolean) flag for request as test of branch vs. branch in git repository
    :param compare_branch_name: (string) the branch name requested by user for test vs. current branch (user specified)
    """

    def __init__(
        self,
        gitrepo_path,
        ufo_directory_list,
        is_commit_test=False,
        commit_number="0",
        is_branch_test=False,
        compare_branch_name=None,
    ):
        self.gitrepo_path = gitrepo_path  # path to root of git repository
        self.ufo_directory_list = ufo_directory_list  # used to filter results by user defined UFO source directory
        self.is_commit_test = is_commit_test  # commit test flag
        self.commit_number = (
            commit_number  # user defined number of commits in git history to compare
        )
        self.is_branch_test = is_branch_test  # branch test flag
        self.compare_branch_name = (
            compare_branch_name  # comparison branch requested by user (if branch test)
        )
        self.current_branch_name = (
            ""  # defined in _define_and_validate_ufo_diff_lists if branch test
        )
        self.ufo = Ufo()  # Ufo class used for UFO source validations
        self.git = None  # GitPython object (instantiated in class method)
        self.delta_fp_string_dict = DeltaFilepathStringDict(
            ufo_directory_list
        )  # stores delta file strings in .delta_dict attribute

        self.commit_sha1_list = (
            []
        )  # used to create dictionaries of report data in class methods

        # file string lists
        self.added_ufo_file_list = (
            []
        )  # used to create dictionaries of report data in class methods
        self.modified_ufo_file_list = (
            []
        )  # used to create dictionaries of report data in class methods
        self.deleted_ufo_file_list = (
            []
        )  # used to create dictionaries of report data in class methods

        # filters files for UFO spec and includes only diff files that are part of spec
        self._define_and_validate_ufo_diff_lists()  # defines many of the class properties on object instantiation

    # PRIVATE METHODS
    def _define_and_validate_ufo_diff_lists(self):
        """
        Defines Delta class properties on instantiation of the object in app.py module
        :return: no return object
        """
        # instantiate git Repo object
        repo = Repo(self.gitrepo_path)
        # define class attribute git object
        self.git = repo.git

        if self.is_commit_test is True:
            commit_number_string = "HEAD~" + self.commit_number  # with HEAD~N syntax
            added_file_string = self.git.diff(
                "--name-only", "--diff-filter=A", commit_number_string
            )
            added_filepath_list = added_file_string.split("\n")
            deleted_file_string = self.git.diff(
                "--name-only", "--diff-filter=D", commit_number_string
            )
            deleted_filepath_list = deleted_file_string.split("\n")
            modified_file_string = self.git.diff(
                "--name-only", "--diff-filter=M", commit_number_string
            )
            modified_filepath_list = modified_file_string.split("\n")
        elif self.is_branch_test is True:
            self.current_branch_name = self.git.rev_parse(["--abbrev-ref", "HEAD"])
            branch_comparison_string = (
                self.compare_branch_name + ".." + self.current_branch_name
            )
            added_file_string = self.git.diff(
                ["--name-only", "--diff-filter=A", branch_comparison_string]
            )
            added_filepath_list = added_file_string.split("\n")
            deleted_file_string = self.git.diff(
                ["--name-only", "--diff-filter=D", branch_comparison_string]
            )
            deleted_filepath_list = deleted_file_string.split("\n")
            modified_file_string = self.git.diff(
                ["--name-only", "--diff-filter=M", branch_comparison_string]
            )
            modified_filepath_list = modified_file_string.split("\n")

        # load class attribute lists with the filepaths that are validated to be UFO in the following method
        self._validate_ufo_and_load_dict_from_filepath_strings(
            added_filepath_list, deleted_filepath_list, modified_filepath_list
        )

    def _add_commit_sha1_to_lists(self):
        """
        Adds commit SHA1 short codes for commits requested by user to the Delta.commit_sha1_list property as a Python
        list.

        :return: no return object
        """
        sha1_num_commits = "-" + self.commit_number
        sha1_args = [sha1_num_commits, "--pretty=%h"]
        sha1_string = self.git.log(
            sha1_args
        )  # git log -[N] --pretty=%h ===> newline delimited list of SHA1 x N commit
        self.commit_sha1_list = sha1_string.split(
            "\n"
        )  # do not modify to os.linesep, Win fails tests with this change

    def _validate_ufo_and_load_dict_from_filepath_strings(
        self, added_filepath_list, deleted_filepath_list, modified_filepath_list
    ):
        """
        Validates filepaths for detected added, modified, and deleted files against UFO specification and includes
        valid files in Delta object property lists.  These lists are used to generate the DeltaFilepathStringDict
        delta_dict Python dictionary that maintains these data in key:value format for report generation.

        UFO validation of files occurs in this class method.

        :param added_filepath_list: (list) files that were added to repository
        :param deleted_filepath_list: (list) files that were deleted from repository
        :param modified_filepath_list: (list) files that were modified in repository
        :return: no return object
        """
        # test for valid UFO files and add the filepath string to the appropriate class instance attribute
        # load added files list
        if len(added_filepath_list) > 0:
            for added_file in added_filepath_list:
                if self.ufo.validate_file(added_file) is True:
                    self.added_ufo_file_list.append(added_file)

        # load deleted files list
        if len(deleted_filepath_list) > 0:
            for deleted_file in deleted_filepath_list:
                if self.ufo.validate_file(deleted_file) is True:
                    self.deleted_ufo_file_list.append(deleted_file)

        # load modified files list
        if len(modified_filepath_list) > 0:
            for modified_file in modified_filepath_list:
                if self.ufo.validate_file(modified_file) is True:
                    self.modified_ufo_file_list.append(modified_file)

        # define the key:value structure of the dictionary attribute on the DeltaFilepathStringDict() class
        if self.is_commit_test:
            self._add_commit_sha1_to_lists()
            self.delta_fp_string_dict.add_commit_sha1(
                self.commit_sha1_list
            )  # create 'commits' dict key
        elif self.is_branch_test:
            branch_list = [self.compare_branch_name, self.current_branch_name]
            self.delta_fp_string_dict.add_branches(
                branch_list
            )  # create 'branches' dict key
        self.delta_fp_string_dict.add_added_filepaths(
            self.added_ufo_file_list
        )  # create 'added' dict key
        self.delta_fp_string_dict.add_deleted_filepaths(
            self.deleted_ufo_file_list
        )  # create 'deleted' dict key
        self.delta_fp_string_dict.add_modified_filepaths(
            self.modified_ufo_file_list
        )  # create 'modified' dict key

    def _get_delta_text_string(self):
        """
        Generates plain text string format for delta subcommand reports.
        :return: (string) plain text string formatted Python string intended for standard output stream
        """
        textstring = ""
        if (
            self.is_commit_test is True
        ):  # include commits if this is an analysis of commit history
            # Write SHA1 commits under examination
            if len(self.delta_fp_string_dict.delta_dict["commits"]) > 0:
                textstring += (
                    os.linesep + "Commit history SHA1 for this analysis:" + os.linesep
                )
                for sha1_commit in self.delta_fp_string_dict.delta_dict["commits"]:
                    textstring += " " + sha1_commit + os.linesep
                textstring += os.linesep
        elif (
            self.is_branch_test is True
        ):  # include branches if this is a branch v branch analysis
            if len(self.delta_fp_string_dict.delta_dict["branches"]) > 0:
                textstring += os.linesep + "Branches under analysis:" + os.linesep
                for branch in self.delta_fp_string_dict.delta_dict["branches"]:
                    textstring += " " + branch + os.linesep
                textstring += os.linesep

        # include added files
        if len(self.delta_fp_string_dict.delta_dict["added"]) > 0:
            for added_file in self.delta_fp_string_dict.delta_dict["added"]:
                add_append_string = "[A]:" + added_file + os.linesep
                textstring += add_append_string
        # include deleted files
        if len(self.delta_fp_string_dict.delta_dict["deleted"]) > 0:
            for deleted_file in self.delta_fp_string_dict.delta_dict["deleted"]:
                del_append_string = "[D]:" + deleted_file + os.linesep
                textstring += del_append_string
        # include modified files
        if len(self.delta_fp_string_dict.delta_dict["modified"]) > 0:
            for modified_file in self.delta_fp_string_dict.delta_dict["modified"]:
                mod_append_string = "[M]:" + modified_file + os.linesep
                textstring += mod_append_string

        return textstring

    def _get_delta_json_string(self):
        """
        Generates JSON format for deltajson subcommand reports.
        :return: (string) JSON formatted Python string intended for standard output stream
        """
        return json.dumps(self.delta_fp_string_dict.delta_dict)

    def _get_delta_markdown_string(self):
        """
        Generates Markdown format for deltamd subcommand reports.
        :return: (string) Markdown formatted Python string intended for standard output stream
        """
        markdown_string = ""

        if self.is_commit_test is True:
            if len(self.delta_fp_string_dict.delta_dict["commits"]) > 0:
                markdown_string += (
                    os.linesep
                    + "## Commit history SHA1 for this analysis:"
                    + os.linesep
                )
                for sha1_commit in self.delta_fp_string_dict.delta_dict["commits"]:
                    markdown_string += "- `" + sha1_commit + "`" + os.linesep
                markdown_string += os.linesep
        elif self.is_branch_test is True:
            if len(self.delta_fp_string_dict.delta_dict["branches"]) > 0:
                markdown_string += (
                    os.linesep + "## Branches under analysis:" + os.linesep
                )
                for branch in self.delta_fp_string_dict.delta_dict["branches"]:
                    markdown_string += "- " + branch + os.linesep
                markdown_string += os.linesep

        # Added files block
        markdown_string += "## Added Files" + os.linesep
        if len(self.delta_fp_string_dict.delta_dict["added"]) > 0:
            for added_file in self.delta_fp_string_dict.delta_dict["added"]:
                markdown_string += "- " + added_file + os.linesep
        else:
            markdown_string += "- None" + os.linesep

        # Deleted files block
        markdown_string += os.linesep + os.linesep + "## Deleted Files" + os.linesep
        if len(self.delta_fp_string_dict.delta_dict["deleted"]) > 0:
            for deleted_file in self.delta_fp_string_dict.delta_dict["deleted"]:
                markdown_string += "- " + deleted_file + os.linesep
        else:
            markdown_string += "- None" + os.linesep

        # Modified files block
        markdown_string += os.linesep + os.linesep + "## Modified Files" + os.linesep
        if len(self.delta_fp_string_dict.delta_dict["modified"]) > 0:
            for modified_file in self.delta_fp_string_dict.delta_dict["modified"]:
                markdown_string += "- " + modified_file + os.linesep
        else:
            markdown_string += "- None" + os.linesep

        # Project URL + version footer
        markdown_string += (
            os.linesep
            + os.linesep
            + "---"
            + os.linesep
            + "[ufodiff](https://github.com/source-foundry/ufodiff) v"
            + major_version
            + "."
            + minor_version
            + "."
            + patch_version
        )

        return markdown_string

    # PUBLIC METHODS

    def get_stdout_string(self, write_format=None):
        """
        Called by app.py module with write_format type that is dependent upon the user subcommand request
        :param write_format: (string) options include 'text', 'json', and 'markdown'
        :return: (string) file change report formatted according to write_format parameter
        """
        if write_format == "text":
            return self._get_delta_text_string()
        elif write_format == "json":
            return self._get_delta_json_string()
        elif write_format == "markdown":
            return self._get_delta_markdown_string()

    # TODO: implement glyph only data
    # def get_stdout_string_glyph_only(self):
    #     pass

    # TODO: implement nonglyph only data
    # def get_stdout_string_nonglyph_only(self):
    #     pass


class DeltaFilepathStringDict(object):
    """
    Object that maintains a Python dictionary of filepaths that meet UFO spec and any user defined UFO path filters
    for use in the generation of standard output strings.

    User defined UFO directory path filter occurs here.
    """

    def __init__(self, ufo_directory_list):
        self.delta_dict = {}
        self.ufo_directory_list = ufo_directory_list

    def _filter_and_load_lists(self, filepath_list):
        the_filepath_list = []
        if (
            len(self.ufo_directory_list) == 0
        ):  # no user defined UFO source filters, include all UFO source
            for delta_file in filepath_list:
                the_filepath_list.append(delta_file)
        else:  # user wants the results to be filtered by specific UFO directory(ies)
            for delta_file in filepath_list:
                for ufo_directory_filter_path in self.ufo_directory_list:
                    if ufo_directory_filter_path in delta_file:
                        the_filepath_list.append(delta_file)
        return the_filepath_list

    def add_added_filepaths(self, added_filepath_list):
        self.delta_dict["added"] = self._filter_and_load_lists(added_filepath_list)

    def add_deleted_filepaths(self, deleted_filepath_list):
        self.delta_dict["deleted"] = self._filter_and_load_lists(deleted_filepath_list)

    def add_modified_filepaths(self, modified_filepath_list):
        self.delta_dict["modified"] = self._filter_and_load_lists(
            modified_filepath_list
        )

    def add_commit_sha1(self, commit_sha1_list):
        self.delta_dict["commits"] = self._filter_and_load_lists(commit_sha1_list)

    def add_branches(self, branch_name_list):
        self.delta_dict["branches"] = self._filter_and_load_lists(branch_name_list)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# ====================================================
# Copyright 2018 Christopher Simpkins
# MIT License
# ====================================================

import os

from git import Repo
from ufodiff.utilities.ufo import Ufo


class Diff(object):
    """
    Diff class performs git diff on repository with filters for the UFO specification, cleans diff reports for
    improved readability and prepends branch names to the report for branch vs. branch analyses.  Supports git diff
    reports with ANSI color codes and without ANSI color codes.

    :param gitrepo_path: (string) path to root of git repository
    :param color_diff: (boolean) indicator for request for color diff (True) or uncolored diff (False)
    """

    def __init__(self, gitrepo_path, color_diff=False):
        self.gitrepo_path = gitrepo_path  # root path for git repository
        self.is_color_diff = color_diff  # is request for color diff = True
        self.repo = Repo(self.gitrepo_path)  # GitPython Repo object
        self.git = self.repo.git  # GitPython Repo.git object
        self.ufo = Ufo()  # ufodiff.utilities.ufo.Ufo object
        self.current_branch = self.git.rev_parse(
            "--abbrev-ref", "HEAD"
        )  # current git branch automatically detected

    # PRIVATE METHODS

    def _clean_diff_string(self, dirty_diff_string):
        """
        'Cleans' the raw git diff string to improve readability and eliminate data that was not felt to be warranted in
        ufodiff text diff reports.

        :param dirty_diff_string: (string) the raw git diff string
        :return: (string) the cleaned git diff string
        """
        dirty_diffstring_list = dirty_diff_string.split("\n")
        clean_diff_string = ""
        for a_string in dirty_diffstring_list:
            if a_string.startswith("\x1b[1mdiff --git") or a_string.startswith(
                "diff --git"
            ):
                clean_a_string = a_string.replace("diff --git", "")
                clean_a_string = clean_a_string.replace(" ", os.linesep)
                clean_diff_string += clean_a_string + os.linesep
            elif "100644" in a_string:
                clean_a_string = a_string.replace("100644", "")
                clean_a_string = clean_a_string.replace("mode", "")
                clean_diff_string += clean_a_string + os.linesep
            else:
                clean_diff_string += a_string + os.linesep
        # remove two lead lines from text diffs (unnecessary duplication of the two files)
        if "---" in clean_diff_string and "+++" in clean_diff_string:
            clean_diff_string_list = clean_diff_string.split(os.linesep)
            purged_head_paths_diff_string_list = clean_diff_string_list[3:]
            clean_diff_string = (
                ""  # reset the diff string to empty string and start again
            )
            for line in purged_head_paths_diff_string_list:
                clean_diff_string += line + os.linesep
        return clean_diff_string

    # PUBLIC METHODS

    def get_diff_string_generator(self, git_user_diff_string):
        """
        Creates a Python generator that returns individual diff reports for filepaths that match UFO spec filters.

        Generator used as the creation of diff string across large numbers of *.glif file changes can take time to
        create.
        :param git_user_diff_string: (string) the string provided as third argument in user command (ufodiff diff [arg3])
        :return: (Python generator of strings) iterable list of diff text strings intended for standard output
        """
        ufo_file_list = self.ufo.get_valid_file_filterlist_for_diff()  # valid UFO files
        is_branch_test = False  # default is a test between commits, not a test between branches, modified below

        if git_user_diff_string.startswith("commits:"):
            commits_list = git_user_diff_string.split(":")
            commits_number = commits_list[1]
            diff_arg_string = "HEAD~" + commits_number
        elif git_user_diff_string.startswith("branch:"):
            is_branch_test = True
            diff_arg_string = git_user_diff_string[7:] + ".." + self.current_branch
        else:
            diff_arg_string = git_user_diff_string

        if self.is_color_diff is True:
            for ufo_file in ufo_file_list:
                dirty_diff_string = self.git.diff(
                    diff_arg_string, "--minimal", "--color", "--", ufo_file
                )
                cleaned_diff_string = self._clean_diff_string(dirty_diff_string)
                if (
                    len(cleaned_diff_string) > 1
                ):  # eliminates diff filters that did not include identified files
                    if is_branch_test is True:
                        cleaned_diff_string = (
                            "branch "
                            + diff_arg_string
                            + os.linesep
                            + cleaned_diff_string
                        )
                    yield cleaned_diff_string
        else:
            for ufo_file in ufo_file_list:
                dirty_diff_string = self.git.diff(
                    diff_arg_string, "--minimal", "--", ufo_file
                )
                cleaned_diff_string = self._clean_diff_string(dirty_diff_string)
                if (
                    len(cleaned_diff_string) > 1
                ):  # eliminates diff filters that did not include identified files
                    if (
                        is_branch_test is True
                    ):  # add branch descriptions to the output from the diff
                        cleaned_diff_string = (
                            "branch "
                            + diff_arg_string
                            + os.linesep
                            + cleaned_diff_string
                        )
                    yield cleaned_diff_string
